/* plucker.c */
#include "headers.h"

Pool *Ppool;	    /* plucker line pool */

int p_verbose = 0;
int p_armed = 0;	/* write input? */
int p_randomize = 1;    /* randomize input? */

CONTOUR **g_contours;
int g_ncontours;

/* note: this module calls fortran
   subroutine dsvdc(x,ldx,n,p,s,e,u,ldu,v,ldv,work,job,info)
   integer ldx,n,p,ldu,ldv,job,info
   double precision x(ldx,1),s(1),e(1),u(ldu,1),v(ldv,1),work(1)
   from file dsvd.c
   */

void p_construct();

static float plucker_basis[2][6] = { { 1.0, 0.0, 0.0, 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0, 0.0, 1.0, 0.0 } };
static PLUCKER plucker_zero = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1, -1 };

typedef struct {
  double *contents;         /* the array */
  int nrow;               /* number of rows */
  int ncol;               /* number of columns */
  int size;               /* number of elements allocated.
						   * size >= nrow * ncol
						   */
} *DArray;

#define AREF(a,i,j)     (a)->contents[(i)*(a)->ncol+(j)]

init_plucker()
{
  void snapshot_all_data();
  
  Ppool = new_pool (sizeof (PLUCKER), 1024, "Ppool");
  nd_set_ubailout (snapshot_all_data);
}

void
set_p_verbose (flag)
	 int flag;
{
  p_verbose = flag;
}

float
p_dot (
	   float *U,
	   float *V,
	   int n)
{
  float dot = 0.0;
  
  while (--n >= 0)
	dot += U[n] * V[n];
  return (dot);
}

float_identical (a,b)
	 float a,b;
{
  return (fabs(a-b) < P_TOL);
}

/* true iff scalar multiples */
/* doesn't handle null vec */
float
p_multiples (U, V, n)
	 float *U;
	 float *V;
	 int n;
{
  int k, kmax=0;
  float scale;
  
  /* find biggest elt */
  for (k = 1; k < n; k++)
	if (ABS(U[k]) > ABS(U[kmax]))
	  kmax = k;
  
  if (V[kmax] == 0.0)
	return (0);
  else
	scale = U[kmax]/V[kmax];
  
  for (k = 0; k < n; k++)
	if (!float_identical (U[k], V[k] * scale))
	  return (0);
  
  return (1);
}

float
p_fdot (U, V, n)
	 float *U;
	 float *V;
	 int n;
{
  float dot = 0.0;
  
  while (--n >= 0)
	dot += U[n] * V[n];
  return (dot);
}

void
p_swap (U, V, n)
	 float *U;
	 float *V;
	 int n;
{
  float t;
  
  while (--n >= 0)	{
	t = U[n];
	U[n] = V[n];
	V[n] = t;    
  }
}

/* norm 1.0 */
/* returns 1 iff P is normalizable */
int
p_normalize (P, n)
	 float *P;
	 int n;
{
  float mag=0.0;
  int j,k;
  
  for (k = 0; k < n; k++)
	mag += (P[k] * P[k]);
  
  if (mag < P_TOL)	{
	if (p_verbose)
	  fprintf (stderr, "p_normalize: zero norm\n");
	return (0);
  }
  else    {
	mag = 1.0 / fsqrt (mag);
	for (k = 0; k < n; k++)
	  P[k] *= mag;
	for (k = 0; k < n; k++)
	  if (fabs(P[k]) < P_TOL)
		P[k] = 0.0;
	for (k = 0; k < n; k++)
	  if (fabs(P[k]) == 1.0)  {
		for (j = 0; j < n; j++)
		  if (j != k)
			P[j] = 0.0;
	  }
	return (1);
  }
}

/* normalize ONLY wrt the first n-1 coeffs; scale the nth accordingly */
int
p_special_normalize (
					 float *P,
					 int n)
{
  float mag=0.0;
  int k;
  
  for (k = 0; k < n-1; k++)
	mag += (P[k] * P[k]);
  
  if (mag < P_TOL)	{
	if (p_verbose)
	  fprintf (stderr, "p_special_normalize: zero norm\n");
	return (0);
  }
  
  mag = 1.0 / sqrt (mag);
  for (k = 0; k < n; k++)
	P[k] *= mag;
  
  return (1);
}

/* w = 1.0 */
void
p_project (P, n)
	 float *P;
	 int n;
{
  int k;
  float max;
  
  if (fabs (P[n-1]) >= P_TOL)	{
	for (k = 0; k < n-1; k++)
	  P[k] /= P[n-1];
	P[n-1] = 1.0;
  }
  else    {
	max = 0.0;
	for (k = 0; k < n-1; k++)
	  if (fabs(P[k]) > fabs(max))
		max = P[k];
	for (k = 0; k < n - 1; k++)
	  P[k] /= max;
	P[n-1] = 0.0;
  }
}

/* A = B */
void
p_assign (A, B, n)
	 float *A, *B;
	 int n;
{
  int k;
  
  for (k = 0; k < n; k++)
	A[k] = B[k];
}

/* A = B */
void
p_fdassign (A, B, n)
	 float *A;
	 double *B;
	 int n;
{
  int k;
  
  for (k = 0; k < n; k++)
	A[k] = B[k];
}

/* A = B */
void
p_dassign (
		   double *A,
		   double *B,
		   int n)
{
  int k;
  
  for (k = 0; k < n; k++)
	A[k] = B[k];
}

/* A = B */
void
p_fassign (
		   float *A, float *B,
		   int n)
{
  int k;
  
  for (k = 0; k < n; k++)
	A[k] = B[k];
}

/* A = -A */
void
p_negate (A, n)
	 float *A;
	 int n;
{
  int k;
  
  for (k = 0; k < n; k++)
	A[k] = -A[k];
}

int
p_ideal (P, n)
	 float *P;
	 int n;
{
  return (fabs(P[n-1]) < P_TOL);
}

int
p_undefined (P, n)
	 float *P;
	 int n;
{
  int k, undefined;
  
  undefined = 1;
  for (k = 0; k < n; k++)
	if (fabs(P[k]) >= P_TOL)
	  undefined = 0;
  
  return (undefined);
}

/* yields inner product of two directed plucker lines */
float
p_product (
		   float *P,
		   float *Q)
{
  return (P[0] * Q[4] + P[1] * Q[5] + P[2] * Q[3] 
		  + Q[0] * P[4] + Q[1] * P[5] + Q[2] * P[3]);
}

/* true iff line P incident to Q */
int
p_incident (P,Q)
	 float *P, *Q;
{
  float mag;
  
  mag = p_product (P, Q);
  
  if (fabsf(mag) > P_TOL)  {
	if (p_verbose)
	  fprintf (stderr, "incident: mag = %9.7f\n", mag);
	return (0);
  }
  else
	return (1);
}

/* print a plucker */
p_print (P)
	 float *P;
{
  fprintf (stderr, "P = %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf\n", 
		   P[0], P[1], P[2], P[3], P[4], P[5]);
}

/* return index of first linearly dependent member found, or -1 if no such member */
p_eliminate (LL, n)
	 float *LL[];
	 int n;
{
  int col, row, maxrow, k;
  float max, scale;
  int dependent;
  static float L[32][6];
  static primary[32];	/* does it have a leading non-zero? */
  
  assure (n <= 32);
  /* copy in user vectors */
  for (k = 0; k < n; k++) {
	for (col = 0; col < 6; col++)
	  L[k][col] = *(LL[k] + col);
	primary[k] = FALSE;
	if (p_verbose)
	  p_print (L[k]);
  }
  
  for (col = 0; col < 6; col++)   {
	/* find non-primary row with largest entry in this column */
	max = 0.0;
	maxrow = -1;
	for (row = 0; row < n; row++)
	  if (!primary[row] && fabs(L[row][col]) > fabs(max)) {
		max = L[row][col];
		maxrow = row;
	  }
	
	/* if this column roughly 0, continue */
	if (fabs(max) < P_TOL)
	  continue;
	
	if (p_verbose)
	  fprintf (stderr, "pivoting on col %d with maxrow %d\n", col, maxrow);
	
	/* otherwise pivot on maxrow */
	primary[maxrow] = TRUE;
	for (row = 0; row < n; row++)	{
	  if (primary[row])	{
		if (p_verbose)
		  fprintf (stderr,"PRI row %d: %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf\n",
				   row, L[row][0], L[row][1], L[row][2], L[row][3], L[row][4], L[row][5]);
		continue;
	  }
	  
	  dependent = TRUE;
	  if (L[row][col] != 0.0) {
		scale = L[row][col] / max;
		for (k = 0; k < 6; k++)	    /* start k at col? */
		  if (fabs(L[row][k] -= L[maxrow][k] * scale) >= P_TOL)
			dependent = FALSE;
	  }
	  else
		for (k = 0; k < 6; k++)	    /* start k at col? */
		  if (fabs(L[row][k]) > P_TOL)
			dependent = FALSE;
	  
	  if (p_verbose)
		fprintf (stderr,"piv row %d: %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf\n",
				 row, L[row][0], L[row][1], L[row][2], L[row][3], L[row][4], L[row][5]);
	  
	  if (dependent)
		return (row);
	}
  }
  
  /* nothing dependent; return -1 */
  return (-1);
}

/* classify P w.r.t. Q */
int
p_classify (P,Q)
	 float *P, *Q;
{
  float mag;
  
  mag = p_product (P, Q);
  
  if (mag < -P_TOL)
	return (P_RHAND);
  else if (mag > P_TOL)
	return (P_LHAND);
  else
	return (P_TIGHT);
}

/* true iff P lies on plucker surface */
int
p_realizable(P)
	 float *P;
{
  return (p_incident (P, P));
}

/* given two 3-d homog points, returns ORIENTED plucker line L through them */
void
p_construct (Ax, Ay, Az, Aw, Bx, By, Bz, Bw, L)
	 float Ax, Ay, Az, Aw, Bx, By, Bz, Bw;
	 float L[6];
{
  L[0] = Ax*By - Ay*Bx;
  L[1] = Ax*Bz - Az*Bx;
  L[2] = Ax*Bw - Aw*Bx;
  L[3] = Ay*Bz - Az*By;
  L[4] = Az*Bw - Aw*Bz;
  L[5] = Aw*By - Ay*Bw;
  
  /* check that L lies on the plucker surface */
  /*
    assure (p_realizable (L));
	*/
}

/* from stolfi, primitives for comp geom (dec src #36), p 205 */
/* CAREFUL:
   stolfi uses different coordinate conventions:
   
   A) his convention : P = [wx, wy, xy, wz, xz, yz]
   my convention : P = [xy, xz, xw, yz, zw, wy]
   so	                [         -      -     ]
   
   B) stolfi defines the 'w' of a point as the zeroth (!) entry
   
   C) plucker points are equivalent to homogeneous points
   
   D) plucker hyperplanes have their order and alternating signs reversed
   */

/* plucker A,B,C,D,E,F to stolfi convention */
void
p_tostolfi (B, A)
	 float B[6];
	 float A[6];
{
  A[0] = -B[2];
  A[1] =  B[5];
  A[2] =  B[0];
  A[3] = -B[4];
  A[4] =  B[1];
  A[5] =  B[3];
}

/* plane A,B,C,D to stolfi convention */
void
f_tostolfi (B, A)
	 float B[4];
	 float A[4];
{
  A[0] =  B[2];
  A[1] = -B[1];
  A[2] =  B[0];
  A[3] = -B[3];
}

/* intersect a plucker line A and flat (i.e., plane) B to get point C */
/* C = line A ^ plane B */
void
p_intersect (
			 float A[6],
			 float B[4],
			 float C[4])
{
  float AA[6], BB[4];
  
  /* AA = stolfi's plucker coords for line A */
  p_tostolfi (A, AA);
  
  /* BB = stolfi's plucker coeffs for plane B */
  f_tostolfi (B, BB);
  
  /* stolfi's w (pg. 205) is c0; ours is c3 */
  /* w */ C[3] = AA[0] * BB[2] - AA[1] * BB[1] + AA[3] * BB[0];
  /* x */ C[0] = AA[0] * BB[3] - AA[2] * BB[1] + AA[4] * BB[0];
  /* y */ C[1] = AA[1] * BB[3] - AA[2] * BB[2] + AA[5] * BB[0];
  /* z */ C[2] = AA[3] * BB[3] - AA[4] * BB[2] + AA[5] * BB[1];
}

/* intersect a plucker line A and flat (i.e., plane) B to get point C */
/* C = line A ^ plane B */
int
p_intersect_line_plane (
						float A[6],
						PLANE *B,
						POINT *C)
{
  float Ipt[4];
  
  p_intersect (A, (float *)B, Ipt);
  if (fabsf(Ipt[3]) > 1.0E-7)	{
	C->px = Ipt[0] / Ipt[3];
	C->py = Ipt[1] / Ipt[3];
	C->pz = Ipt[2] / Ipt[3];
	return (1);
  }
  else    {
	C->px = C->py = C->pz = 0.0;
	return (0);
  }
}

/* orient plucker line L so that it goes through plane P then plane Q */
/* may involve changing the signs of each of L's coordinates */
/* plane P and Q are parallel planes */
/* return 0 if L is unorientable (i.e., parallel) w.r.t. P & Q */
int
p_orient (L, P, Q)
	 float L[6], P[4], Q[4];
{
  float A[4], B[4];
  float LL[6], NormL[6];
  int k, neg, pos;
  
  if (!p_realizable (L))  {
	if (p_verbose > 1)
	  fprintf (stderr, "can't orient L: unrealizable line\n");
	return (0);
  }
  
  p_intersect (L, P, A);
  p_intersect (L, Q, B);
  
  /* if either point ideal, unorientable */
  if (p_ideal (A, 4) || p_ideal (B, 4))   {
	if (p_verbose > 1)
	  fprintf (stderr, "L unorientable w.r.t A = %6.3lf %6.3lf %6.3lf %6.3lf, B = %6.3lf %6.3lf %6.3lf %6.3lf\n",
			   A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3]);
	return (0);
  }
  
  /* now reconstruct the ordered line using these two points */
  p_construct (A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3], LL);
  
  /* copy input line into NL, normalize it */
  p_fassign (NormL, L, 6);
  p_normalize (NormL, 6);
  p_normalize (LL, 6);
  
  /* LL must be within a sign (and tolerance) of NL */
  for (pos = neg = k = 0; k < 6; k++) {
	pos += float_identical ( NormL[k], LL[k]);
	neg += float_identical (-NormL[k], LL[k]);
  }
  
  /* now, everything must have matched to within a sign */
  if ((pos != 6) == (neg != 6))   {
#ifdef VERBOSE
	fprintf (stderr, "p_orient(): pos = %d, neg = %d?\n", pos, neg);
	fprintf (stderr, "A = %6.3lf %6.3lf %6.3lf %6.3lf, B = %6.3lf %6.3lf %6.3lf %6.3lf\n",
			 A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3]);
	p_print (LL);
	p_print (NormL);
#endif
	return (0);
  }
  else if (neg == 6)
	p_negate (L, 6);
  
  return (1);
}

static float XYplane[4] = {0.0, 0.0, 1.0, 0.0}; /* x = 0, etc. */
static float YZplane[4] = {1.0, 0.0, 0.0, 0.0};
static float XZplane[4] = {0.0, 1.0, 0.0, 0.0};

static float xyMplane[4] = {0.0, 0.0, 1.0, 1.0}; /* z = -1, etc. */
static float yzMplane[4] = {1.0, 0.0, 0.0, 1.0};
static float xzMplane[4] = {0.0, 1.0, 0.0, 1.0};

static float xyPplane[4] = {0.0, 0.0, 1.0, -1.0}; /* z = +1, etc. */
static float yzPplane[4] = {1.0, 0.0, 0.0, -1.0};
static float xzPplane[4] = {0.0, 1.0, 0.0, -1.0};

/* return two 3-d homog points incident to plucker line L */
void
p_deconstruct (
			   float L[6],
			   float A[4], float B[4])
{
  float sanity[6], Pxy[4], Pyz[4], Pxz[4];
  int xy_undefined, yz_undefined, xz_undefined;
  int xy_ideal, yz_ideal, xz_ideal;
  float *C;
  
  A[0] = A[1] = A[2] = A[3] = 0.0;
  B[0] = B[1] = B[2] = B[3] = 0.0;
  
  p_intersect (L, xyPplane, Pxy);
  p_intersect (L, yzPplane, Pyz);
  p_intersect (L, xzPplane, Pxz);
  
#ifdef CRUFT
  p_project (Pxy, 4);
  p_project (Pyz, 4);
  p_project (Pxz, 4);
#endif
  
  xy_undefined = p_undefined (Pxy, 4);
  yz_undefined = p_undefined (Pyz, 4);
  xz_undefined = p_undefined (Pxz, 4);
  
  xy_ideal = xy_undefined || p_ideal (Pxy, 4);
  yz_ideal = yz_undefined || p_ideal (Pyz, 4);
  xz_ideal = xz_undefined || p_ideal (Pxz, 4);
  
  if (xy_undefined && yz_undefined && xz_undefined)	{
	if (p_verbose)
	  fprintf (stderr, "no determinate intersections for A!\n");
	return;
  }
  
  /* pick A as defined non-ideal; failing that, choose it as ideal */
  if (xy_ideal && yz_ideal && xz_ideal)   {
	if (!xy_undefined)
	  p_fassign (A, Pxy, 4);
	else if (!yz_undefined)
	  p_fassign (A, Pyz, 4);
	else if (!xz_undefined)
	  p_fassign (A, Pxz, 4);
  }
  else    {
#define INSIDE_CUBE 1
#ifdef INSIDE_CUBE
	
	if (!xy_ideal)
	  p_project (Pxy, 4);
	if (!yz_ideal)
	  p_project (Pyz, 4);
	if (!xz_ideal)
	  p_project (Pxz, 4);
	
	C = A;
	
	/* choose one inside unit cube */
	if (!xy_ideal && fabs(Pxy[0]) <= 1.0 && fabs(Pxy[1]) <= 1.0)	{
	  p_fassign (C, Pxy, 4);
	  C = (C == A) ? B : NULL;
	}
	
	if (!yz_ideal && fabs(Pyz[1]) <= 1.0 && fabs(Pyz[2]) <= 1.0)   {
	  p_fassign (C, Pyz, 4);
	  C = (C == A) ? B : NULL;
	}
	
	if (!xz_ideal && fabs(Pxz[0]) <= 1.0 && fabs(Pxz[2]) <= 1.0)   {
	  p_fassign (C, Pxz, 4);
	  C = (C == A) ? B : NULL;
	}
	
	if (C == A) {
	  if (!xy_ideal)
	    p_fassign (A, Pxy, 4);
	  else if (!yz_ideal)
	    p_fassign (A, Pyz, 4);
	  else if (!xz_ideal)
	    p_fassign (A, Pxz, 4);
	  else
	    fprintf (stderr, "deconstruct(): can't place positive point!\n");
	}
#else
	if (!xy_ideal)
	  p_fassign (A, Pxy, 4);
	else if (!yz_ideal)
	  p_fassign (A, Pyz, 4);
	else if (!xz_ideal)
	  p_fassign (A, Pxz, 4);
#endif	/* NOT INSIDE_CUBE */
  }
  
  p_intersect (L, xyMplane, Pxy);
  p_intersect (L, yzMplane, Pyz);
  p_intersect (L, xzMplane, Pxz);
  
#ifdef CRUFT
  p_project (Pxy, 4);
  p_project (Pyz, 4);
  p_project (Pxz, 4);
#endif
  
  xy_undefined = p_undefined (Pxy, 4);
  yz_undefined = p_undefined (Pyz, 4);
  xz_undefined = p_undefined (Pxz, 4);
  
  xy_ideal = xy_undefined || p_ideal (Pxy, 4);
  yz_ideal = yz_undefined || p_ideal (Pyz, 4);
  xz_ideal = xz_undefined || p_ideal (Pxz, 4);
  
  if (xy_undefined && yz_undefined && xz_undefined)	{
	if (p_verbose)
	  fprintf (stderr, "no determinate intersections for B!\n");
	return;
  }
  
  /* pick B as defined non-ideal; failing that, choose it as ideal */
  if (xy_ideal && yz_ideal && xz_ideal)   {
	if (!xz_undefined)
	  p_fassign (B, Pxz, 4);
	else if (!yz_undefined)
	  p_fassign (B, Pyz, 4);
	else if (!xy_undefined)
	  p_fassign (B, Pxy, 4);
  }
  else    {
#ifdef INSIDE_CUBE
	if (!xy_ideal)
	  p_project (Pxy, 4);
	if (!yz_ideal)
	  p_project (Pyz, 4);
	if (!xz_ideal)
	  p_project (Pxz, 4);
	
	if (C == NULL)
	  return;
	
	/* choose one inside unit cube */
	if (!xy_ideal && fabs(Pxy[0]) <= 1.0 && fabs(Pxy[1]) <= 1.0)	{
	  if (C) p_fassign (C, Pxy, 4);
	  C = (C == A) ? B : NULL;
	}
	
	if (!yz_ideal && fabs(Pyz[1]) <= 1.0 && fabs(Pyz[2]) <= 1.0)   {
	  if (C) p_fassign (C, Pyz, 4);
	  C = (C == A) ? B : NULL;
	}
	
	if (!xz_ideal && fabs(Pxz[0]) <= 1.0 && fabs(Pxz[2]) <= 1.0)   {
	  if (C) p_fassign (C, Pxz, 4);
	  C = (C == A) ? B : NULL;
	}
    
	if (C != NULL)	{
	  if (!xy_ideal)
	    p_fassign (B, Pxy, 4);
	  else if (!yz_ideal)
	    p_fassign (B, Pyz, 4);
	  else if (!xz_ideal)
	    p_fassign (B, Pxz, 4);
	  else
	    fprintf (stderr, "deconstruct(): can't place positive point!\n");
	}
#else
	if (!xz_ideal)
	  p_fassign (B, Pxz, 4);
	else if (!yz_ideal)
	  p_fassign (B, Pyz, 4);
	else if (!xy_ideal)
	  p_fassign (B, Pxy, 4);
#endif
  }
  
  p_project (A, 4);
  p_project (B, 4);
  /*
    p_construct (A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3], sanity);
    assure (p_realizable (sanity));
	*/
}

int
p_finiteline (L)
	 float L[6];
{
  float A[4], B[4];
  
  p_deconstruct (L, A, B);
  if (fabs(A[3]) < P_TOL && fabs(B[3]) < P_TOL)
	return (0);
  else return (1);
}

static int perm[6] = { 4, 5, 3, 2, 0, 1 };

/* p_incident4(): return the line T incident to P,Q,R, and S */
/* returns # of roots (0, 1, 2), or:
   
   3 if a solution is a plucker line embedded in plucker surface
   T1 and T2 are the spanning basis
   6 + n
   if (n-2) vectors are linearly dependent 
   T1 and T2 are tight on the inputs
   */
int
p_incident4 (P, Q, R, S, T1, T2)
	 float *P, *Q, *R, *S, *T1, *T2;
{
  int j, k, info;
  double M[6][6], U[6][6], V[6][6];
  double work[6], svalues[6], e[6];
  static int six = 6;
  static int eleven = 11;
  double ill_condition;
  double a, b, c, det, mag; /* for solving, classifying quadratic */
  double s1,t1, s2,t2;      /* homog roots 1 and 2 */
  double *F, *G;
  int nsolutions = 0, rank;
  
  if (p_verbose > 1)  {	
	fprintf (stderr, "\tp_incident:\n");
	p_print (P);
	p_print (Q);
	p_print (R);
	p_print (S);
  }
  
  /* load M, fortran-style (column major) */
  for (j = 0; j < 6; j++) {
	M[j][0] = P[perm[j]];
	M[j][1] = Q[perm[j]];
	M[j][2] = R[perm[j]];
	M[j][3] = S[perm[j]];
	M[j][4] = 0.0;
	M[j][5] = 0.0;
  }
  
  dsvdc_ (M, &six, &six, &six, svalues, e, U, &six, V, &six, work, &eleven, &info);
  
  if (info != 0)  {
	fprintf (stderr, "svd failed, info = %d\n", info);
	return (0);
  }
  else {
	rank = 0;
	for (j = 0; j < 6; j++) {
	  if (fabs(svalues[j]) > P_TOL)
		rank++;
	}
#ifdef VERBOSE
	fprintf (stderr, "svd rank %d (sigma = %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf %6.3lf)\n", 
		     rank, svalues[0], svalues[1], svalues[2], svalues[3], svalues[4], svalues[5]);
	fprintf (stderr, "so rank of intersection space is 5 - %d == %d\n", rank, 5 - rank);
#endif
  }
  
  /* pick out, from V, spanning basis of M's null space (last two columns of V) */
  F = V[4];
  G = V[5];
  
#ifdef CRUFT
  ill_condition = fabs(svalues[0] / svalues[3]);
  if (fabs(svalues[3]) < P_TOL || ill_condition > 1000.0)  {
	if (p_verbose)
	  fprintf (stderr, "svd failed, ill condition (sigma = %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf)\n", 
			   svalues[0], svalues[1], svalues[2], svalues[3], svalues[4], svalues[5]);
	
	for (k = 0; k < 6; k++)
	  if (svalues[k] < P_TOL || (fabs(svalues[0] / svalues[k]) > 1000.0)) {
		nsolutions++;
		
		if (p_finiteline(V[k]))	{
		  if (!p_finiteline(F))
			F = V[k];
		  else if (!p_finiteline(G))
			G = V[k];
		}
	  }
	
	if (p_verbose)
	  fprintf (stderr, "p_incident (degenerate) returning %d\n", 6 + nsolutions);
	
	p_fdassign (T1, F, 6);
	p_fdassign (T2, G, 6);
	
	assure (p_realizable (T1));
	assure (p_realizable (T2));
	/*
	  assure (p_incident (P, T1));
	  assure (p_incident (Q, T1));
	  assure (p_incident (R, T1));
	  assure (p_incident (S, T1));
	  
	  assure (p_incident (P, T2));
	  assure (p_incident (Q, T2));
	  assure (p_incident (R, T2));
	  assure (p_incident (S, T2));
	  */
	return (6 + nsolutions);
  }
#endif
  
  /* find the root of [tF + G] \odot [tF + G] = 0 */
  a = b = c = 0.0;
  for (k = 0; k < 6; k++) {
	a += F[k] * F[perm[k]];
	b += F[k] * G[perm[k]];
	c += G[k] * G[perm[k]];
  }
  
  mag = a * a + b * b + c * c;
  
#ifdef VERBOSE
  fprintf ( stderr, "a, b, c, mag are %4.3f %4.3f %4.3f %4.3f\n", a, b, c, mag );
#endif
  
  if (fabs(mag) < P_TOL)  {
#ifdef VERBOSE
	fprintf ( stderr, "a = b = c = 0; all t are solutions!\n" );
#endif
	nsolutions = 3;
	s1 = t2 = 1.0;
	s2 = t1 = 0.0;
  }
  else    {
	mag = 1.0 / fsqrt(mag);
	a *= mag;
	b *= mag;
	c *= mag;
	
	det = b * b - a * c;	
#ifdef VERBOSE
	fprintf ( stderr, "det = b^2 - ac is %4.3f\n", det );
#endif
	
	/* assign homogenous roots.  note that equation tF + G = 0 is homogenous */
	/* so any non-zero scalar multiple can be applied to it.  in particular */
	/* if t = b/a and a != 0 (for example), then bF + aG = 0 is a solution */
	if (det < 0.0)  {
#ifdef VERBOSE
	  fprintf ( stderr, "det < 0; no real solutions!\n" );
#endif
#define IMAGINARYLINES
#ifdef IMAGINARYLINES
	  if (p_verbose)
		fprintf (stderr, "p_incident4() failed, no real roots (det = %f)\n", det);
	  return (0);
#else
	  /* construct a line through three of the inputs, oriented w.r.t fourth */
	  p_fdassign (T1, F, 6);
	  p_fdassign (T2, G, 6);
	  return (2);
#endif
	}
	else if ( det < P_TOL )	{	/* b^2 roughly equal to a c */
	  /* b^2 - ac ~= 0 */
#ifdef VERBOSE
	  fprintf ( stderr, "det < tolerance; b^2 = ac; only 1 solution!\n" );
#endif
	  nsolutions = 1;
	  
	  if (fabs(a) > fabs(c))  {
		/* multiply root t = <-b/a,1> by a */
		/* to attain root s1 = -b, t1 = a */
        s1 = s2 = -b; 
		t1 = t2 =  a;
	  }
	  else    {
		/* MIKE:  what is the derivation of this root? */
		/* to attain root s1 = c, t1 = -b */
		s1 = s2 =  c;
		t1 = t2 = -b;
	  }
    }
	else	{
#ifdef VERBOSE
	  fprintf ( stderr, "det generic; two solutions\n" );
#endif
	  nsolutions = 2;
	  det = sqrt(det);

	  /* MIKE: why the dependence on the sign of b? */
	  if (b < 0.0)    {
		s1 = -b + det;
		t1 = a;
		/* MIKE: and why is this root <c, -b + det> */
		s2 = c;
		t2 = -b + det;
	  }
	  else    {
		s1 = -b - det;
		t1 = a;
		/* MIKE: ditto here -- shouldn't roots be -b +/- det */
		s2 = c;
		t2 = -b - det;
	  }
	}
  }
  
  /* construct the solutions by linear interpolation */
  for (k = 0; k < 6; k++) {
	T1[k] = s1 * F[k] + t1 * G[k];
	T2[k] = s2 * F[k] + t2 * G[k];
  }
  
  /* normalize the solutions */
  p_normalize (T1, 6);
  if (nsolutions > 1)
	p_normalize (T2, 6);
  
  /* the two solutions should lie on the plucker surface */
  assure (p_realizable (T1));
  if (nsolutions > 1)
	assure (p_realizable (T2));
  
  if (!(p_realizable(T1) && p_realizable(T2)))    {
	int opv = p_verbose;
	
	p_verbose = 1;
	p_realizable (T1);	p_print (T1);
	p_realizable (T2);	p_print (T2);
	p_verbose = opv;
  }
  
  /* verify that the solutions are incident with each of the input lines */
  assure (p_incident (P, T1));
  assure (p_incident (Q, T1));
  assure (p_incident (R, T1));
  assure (p_incident (S, T1));
  
  if (nsolutions > 1)	{
	assure (p_incident (P, T2));
	assure (p_incident (Q, T2));
	assure (p_incident (R, T2));
	assure (p_incident (S, T2));
  }
  
  if (p_verbose)
	fprintf (stderr, "p_incident4 returning %d solutions\n", nsolutions);
  
  /*
    if (nsolutions == 1)    {
	fprintf (stderr, "p_incident4 has 1 solution!\n");
	fprintf (stderr, "a = %lf, b = %lf, c = %lf, mag = (%lf) %lf, det = %lf\n", a, b, c, 1.0/mag, mag, det);
	fprintf (stderr, "s1 = %lf, t1 = %lf, s2 = %lf, t2 = %lf\n", s1, t1, s2, t2);
	fprintf (stderr, "V[2], V[3], V[4], V[5]:\n");
	p_print (V[2]);
	p_print (V[3]);
	p_print (V[4]);
	p_print (V[5]);
	fprintf (stderr, "T1, T2:\n");
	p_print (T1);
	p_print (T2);
	}
	*/
  
  return (nsolutions);
}

void
init_random()
{
  int k;
  /*
    k = getpid ();
    while (k-- > 0)
	(void)rand();
	*/
}

#ifdef PUNDERLINE

/* pl_incident3(): return a line L incident to three pluckers P, Q, and R */
/* uses global contours */
/* finds line through R and plane of P's midpoint with segment Q */
int
pl_incident3 (P, Q, R, L)
	 PLUCKER *P, *Q, *R;
	 float *L;
{
  POINT M_M, *M = &M_M;
  POINT A_A, *A = &A_A;
  POINT B_B, *B = &B_B;
  float N[4];
  float DP[4], DN[4];
  
  /* midpoint of segment 0 */
  M->px = AVG(contours[P->uid][P->pid].px, CX(contours[P->uid], (P->pid + 1) % CLENGTH(contours[P->uid])));
  M->py = AVG(contours[P->uid][P->pid].py, CY(contours[P->uid], (P->pid + 1) % CLENGTH(contours[P->uid])));
  M->pz = AVG(contours[P->uid][P->pid].pz, CZ(contours[P->uid], (P->pid + 1) % CLENGTH(contours[P->uid])));
  
  /* first point of segment 1 */
  A->px = contours[Q->uid][Q->pid].px;
  A->py = contours[Q->uid][Q->pid].py;
  A->pz = contours[Q->uid][Q->pid].pz;
  
  /* second point of segment 1 */
  B->px = CX(contours[Q->uid], (Q->pid + 1) % CLENGTH(contours[Q->uid]));
  B->py = CY(contours[Q->uid], (Q->pid + 1) % CLENGTH(contours[Q->uid]));
  B->pz = CZ(contours[Q->uid], (Q->pid + 1) % CLENGTH(contours[Q->uid]));
  
  /* plane through M and segment 1 */
  if (cross3f (B->px - A->px, B->py - A->py, B->pz - A->pz,
			   A->px - M->px, A->py - M->py, A->pz - M->pz, N) == 0)
	p_construct (M->px, M->py, M->pz, 1.0,
				 AVG(CX(contours[R->uid],R->pid), CX(contours[R->uid], (R->pid + 1) % CLENGTH(contours[R->uid])));
				 AVG(CY(contours[R->uid],R->pid), CY(contours[R->uid], (R->pid + 1) % CLENGTH(contours[R->uid])));
				 AVG(CZ(contours[R->uid],R->pid), CZ(contours[R->uid], (R->pid + 1) % CLENGTH(contours[R->uid])));
				 1.0,
				 L);
  else	{
	N[3] = -(N[0] * A->px + N[1] * A->py + N[2] * A->pz);
	DN[0] = N[0]; DN[1] = N[1]; DN[2] = N[2]; DN[3] = N[3];
	p_intersect (R->L, DN, DP);
	p_construct ((float)M->px, (float)M->py, (float)M->pz, 1.0, DP[0], DP[1], DP[2], DP[3], L);
  }
  
  p_normalize (L, 6);
  
  /* sanity check */
  assure (p_realizable (L));
  assure (p_finiteline (L));
  assure (p_incident (L, P->L));
  assure (p_incident (L, Q->L));
  assure (p_incident (L, R->L));
  
  return (1);
}

p_list_orient (L, slack)
	 double L[6];
	 ARRAY *slack;
{
  int k;
  PLUCKER *S;
  
  for (k = 0; k < ALENGTH(slack); k++)    {
	S = (PLUCKER *) select_array (slack, k);
	if (p_classify (L, S->L) == P_RHAND)   {
	  p_negate (L, 6);
	  break;
	}
  }
}

/*
  find a stabbing line through n_p oriented planar polygons 
  having n edges
  
  based on the linear time smallest enclosing ball algorithm
  */

/* invariant: returns at least one line in L1 or L2 that satisfies constraints */
/* if returns 0, there is no satisfying line.  if returns >= 3, there are infinity */
int
p_underline (slack, nslack, tight, ntight, L1, L2, ref1, ref2, proof, broken)
	 ARRAY *slack;	    /* constraints */
	 int nslack;
	 ARRAY *tight;
	 int ntight;
	 double L1[6];	    /* 0,1, or 2 solutions */
	 double L2[6];
	 double ref1[4];	    /* reference planes */
	 double ref2[4];
	 ARRAY *proof;	    /* five failure lines */
	 ARRAY *broken;	    /* broken constraint lines */
{
  PLUCKER *C[5], *S, *T;
  int k, s, nsolutions;
  int or1, or2, ok1, ok2;
  ARRAY *subtight;
  
  if (p_verbose)  {
	fprintf (stderr, "entering p_underline: |s| == %d, |t| == %d\n", nslack, ntight);
	fprintf (stderr, "  tight originating segments:");
	for (k = 0; k < ntight; k++)	{
	  C[k] = (PLUCKER *) select_array (tight, k);
	  fprintf (stderr, "\t%d, %d;", C[k]->uid, C[k]->pid);
	}
	fprintf (stderr, "\n");
  }
  
  /* should never have more than four tight constraints */
  assert (ntight <= 4);
  
  /* assemble all tights, and as many slacks as needed, to get 4 constraints */
  for (k = 0; k < ntight; k++)
	C[k] = (PLUCKER *) select_array (tight, k);
  for (k = ntight; k < MIN(ntight + nslack, 4); k++)
	C[k] = (PLUCKER *) select_array (slack, k - ntight);
  for (k = MIN(nslack + ntight, 4); k < 4; k++)
	C[k] = &plucker_zero;
  
  if ((nslack > 0 && ntight == 0) || (ntight >= 4)) {
	nsolutions = p_incident4 (C[0]->L, C[1]->L, C[2]->L, C[3]->L, L1, L2);
	
	/* exception case: imaginary solution (no lines through four tight inputs) */
	if (ntight == 4 && nsolutions == 0)	{
	  for (k = 0; k < 4; k++)	{
		pl_incident3 (C[(k + 1) % 4], C[(k + 2) % 4], C[(k + 3) % 4], L1);
		if (p_finiteline (L1) && p_orient (L1, ref1, ref2) && (p_classify (L1, C[k]->L) != P_RHAND))
		  break;
	  }
	  if (k < 4)	{
		nsolutions = 1;
		ok1 = TRUE;
		ok2 = FALSE;
	  }
	  else    {
		fprintf (stderr, "ntight == 4, couldn't find oriented non-imaginary solution\n");
		for (k = 0; k < 4; k++)
		  insert_array (proof, select_array (tight, k));
		return (-1);
	  }
	}
  }
  else if (ntight == 3)   {	    /* line through three originating segments */
	pl_incident3 (C[0], C[1], C[2], L1);
	pl_incident3 (C[1], C[2], C[0], L2);
	p_normalize (L1, 6);
	p_normalize (L2, 6);
	nsolutions = 2;
  }
  else if (ntight == 2)   {	    /* line through midpoints of two originating segments */
	p_construct (AVG(CX(contours[C[0]->uid], C[0]->pid),
					 CX(contours[C[0]->uid], (C[0]->pid + 1) % lengths[C[0]->uid])),
				 AVG(CY(contours[C[0]->uid][C[0]->pid]),
					 CY(contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]])),
				 AVG(CZ(contours[C[0]->uid][C[0]->pid]),
					 CZ(contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]])),
				 1.0,
				 AVG(CX(contours[C[1]->uid], C[1]->pid),
					 CX(contours[C[1]->uid], (C[1]->pid + 1) % CLENGTH(contours[C[1]->uid]))),
				 AVG(CY(contours[C[1]->uid], C[1]->pid),
					 CY(contours[C[1]->uid], (C[1]->pid + 1) % CLENGTH(contours[C[1]->uid]].py),
					 AVG(CZ(contours[C[1]->uid],
							[C[1]->pid].pz, contours[C[1]->uid][(C[1]->pid + 1) % lengths[C[1]->uid]].pz),
						 1.0,
						 L1);
					 p_normalize (L1, 6);
					 nsolutions = 1;
					 }
				 else if (ntight == 1)   {	    /* line through midpoint of originating segment, normal to reference plane */
				   p_construct (AVG(contours[C[0]->uid][C[0]->pid].px, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].px),
								AVG(contours[C[0]->uid][C[0]->pid].py, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].py),
								AVG(contours[C[0]->uid][C[0]->pid].pz, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].pz),
								1.0,
								AVG(contours[C[0]->uid][C[0]->pid].px, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].px) + ref1[0],
								AVG(contours[C[0]->uid][C[0]->pid].py, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].py) + ref1[1],
								AVG(contours[C[0]->uid][C[0]->pid].pz, contours[C[0]->uid][(C[0]->pid + 1) % lengths[C[0]->uid]].pz) + ref1[2],
								1.0,
								L1);
				   p_normalize (L1, 6);
				   nsolutions = 1;
				 }
    else
	assert (FALSE);	    /* not reached */	

#define OLDWAY
#ifdef OLDWAY
    if (ok1 = (nsolutions >= 1) && p_finiteline (L1))
        or1 = p_orient (L1, ref1, ref2);
    if (ok2 = (nsolutions >= 2) && p_finiteline (L2))
        or2 = p_orient (L2, ref1, ref2);
/*
    if (ok1 && !or1)
	p_list_orient (L1, slack);
    if (ok2 && !or2)
	p_list_orient (L2, slack);
*/
#else
    if (ok1 = (nsolutions >= 1))
	if (p_finiteline (L1))
            p_orient (L1, ref1, ref2);
    if (ok2 = (nsolutions >= 2))
	if (p_finiteline (L2))
	    p_orient (L2, ref1, ref2);
#endif

    /* must satisfy all remaining slack constraints */
    for (s = 0; s < nslack; s++)    {
	S = (PLUCKER *) select_array (slack, s);
	if (ok1)
	    ok1 &= (p_classify (L1, S->L) != P_RHAND);
	if (ok2)
	    ok2 &= (p_classify (L2, S->L) != P_RHAND);

	if (!(ok1 || ok2))  {	/* S failed; generate & solve subinstance with S tight */
	    if (ntight == 4)	{	    /* no solution */
		if (!p_finiteline(L1) && !p_finiteline(L2))
		    fprintf (stderr, "p_und failing on two infinite lines!\n");
		for (k = 0; k < 4; k++)
		    insert_array (proof, select_array (tight, k));
		insert_array (proof, (void *)S);

	        insert_array (broken, (void *)L1);
		if (nsolutions > 1)
		    insert_array (broken, (void *)L2);

		return (0);
		}

	    /* solve subinstance to get (oriented) solutions L1 & L2 */
            subtight = copy_array (tight);
	    /* subinstance is tight on the newly broken constraint S */
	    insert_array (subtight, (void *)S);
	    nsolutions = p_underline (slack, s, subtight, ntight + 1, L1, L2, ref1, ref2, proof, broken);

	    if (nsolutions <= 0)    {
	        discard_array (&subtight);
		return (nsolutions);
		}

	    /* check satisfaction of remaining constraints */
	    if (ok1 = (nsolutions >= 1))
		if (p_finiteline (L1))
		    or1 = p_orient (L1, ref1, ref2);
	    if (ok2 = (nsolutions >= 2))
		if (p_finiteline (L2))
		    or2 = p_orient (L2, ref1, ref2);
/*
	    if (ok1 && !or1)
		p_list_orient (L1, slack);
	    if (ok2 && !or2)
		p_list_orient (L2, slack);
*/

	    assert (ok1 || ok2);

	    /* sanity check : does some solution satisfy everything we called p_() on? */
	    for (k = 0; k <= s; k++)  {
		S = (PLUCKER *) select_array (slack, k);
		if ((p_classify (L1, S->L) == P_RHAND)
		    && (p_classify (L2, S->L) == P_RHAND))
		    fprintf (stderr, "p_und [SLACK]: ns, nt, %d, %d; nsubslack, nsubtight %d, %d, k == %d\n", 
			nslack, ntight, s, ntight + 1, k);
		}

	    for (k = 0; k < ntight + 1; k++)  {
		T = (PLUCKER *) select_array (subtight, k);
		if ((p_classify (L1, T->L) == P_RHAND)
		    && (p_classify (L2, T->L) == P_RHAND))
		    fprintf (stderr, "p_und [TIGHT]: ns, nt, %d, %d; nsubslack, nsubtight %d, %d, k == %d\n",
			nslack, ntight, s, ntight + 1, k);
		}

	    discard_array (&subtight);
	    }	/* recursive case */
	}   /* for s */

    /* sanity check : does some solution satisfy everything we called p_() on? */
    for (k = 0; k < nslack; k++)  {
	S = (PLUCKER *) select_array (slack, k);
	if ((p_classify (L1, S->L) == P_RHAND)
	    && (p_classify (L2, S->L) == P_RHAND))
	    fprintf (stderr, "p_und(): [END SLACK]: ns, nt, %d, %d; k == %d\n", nslack, ntight, k);
	}

    for (k = 0; k < ntight; k++)  {
	T = (PLUCKER *) select_array (tight, k);
	if ((p_classify (L1, T->L) == P_RHAND)
	    && (p_classify (L2, T->L) == P_RHAND))
	    fprintf (stderr, "p_und(): [END TIGHT]: ns, nt, %d, %d; k == %d\n", nslack, ntight, k);
	}

    if (!ok1 && ok2)
	p_dassign (L1, L2, 6);

    if (!(ok1 || ok2))	{
	assure (ok1 || ok2);
	fprintf (stderr, "nslack == %d, ntight == %d\n", nslack, ntight);
	fprintf (stderr, "nsolns == %d, L1%sfinite, L2%sfinite:\n",
		nsolutions, p_finiteline(L1) ? " " : " NOT ", p_finiteline(L2) ? " " : " NOT ");
	fprintf (stderr, "L1%srealizable, L2%srealizable:\n", 
		p_realizable(L1) ? " " : " NOT ", p_realizable(L2) ? " " : " NOT ");
	p_print (L1);
	p_print (L2);
	for (k = 0; k < MIN(4, nslack + ntight); k++)
	    insert_array (proof, C[k]);
	return (0);
	}

    return (ok1 + ok2);
}

/* super_p_stabbers():
    try hard to return stabbing line(s) through a collection of n oriented polygons */
int
super_p_stabbers (
    CONTOUR **contours,
    int n,			    /* # of contours */
    float *L1, float *L2,	    /* stabbing lines if return value > 0 */	
    ARRAY *proof,		    /* proof of failure if nsolns == 0 */
    ARRAY *broken,	    	    /* broken constraints if applicable */
    int *attempts)
{
ARRAY *slack, *tight, *perm_slack;
int k, len, cid, pid, nsolns;
POINT Ca, Cb, P_P, *P = &P_P, Q_Q, *Q = &Q_Q;
PLUCKER PL, *PP;
float ref1[4], ref2[4];
static int oldslack = -1;
CONTOUR *c;

    slack = create_array (sizeof (PLUCKER));
    tight = create_array (sizeof (PLUCKER));

    for (cid = 0; cid < n; cid++) {
	c = contours[cid];
	if (CLENGTH(c) >= 2)
            for (pid = 0; pid < CLENGTH(c); pid++)	{
	        /* construct the oriented plucker line from vk to vk+1 */
		P = CVERTEX(contours[cid], pid);
		Q = CVERTEX(contours[cid], (pid + 1) % CLENGTH(c));
		PL.uid = cid;
		PL.pid = pid;
		p_construct (P->px, P->py, P->pz, 1.0, 
			     Q->px, Q->py, Q->pz, 1.0, PL.L);
		p_normalize (PL.L, 6);
		insert_array (slack, (void *)&PL);

	        /* special case: if this is a length 2 contour, don't enter symmetric constraint */
		if (CLENGTH(c) == 2)
		    break;
		}

	}

    /* generate reference plane normal to vector from centroid 0 to n */
    get_ccentroid (    0, &Ca.px, &Ca.py, &Ca.pz); 
    get_ccentroid (n - 1, &Cb.px, &Cb.py, &Cb.pz);

    ref1[0] = Cb.px - Ca.px;
    ref1[1] = Cb.py - Ca.py;
    ref1[2] = Cb.pz - Ca.pz;
    p_normalize (ref1, 3);
    ref1[3] = -(ref1[0] * Ca.px + ref1[1] * Ca.py + ref1[2] * Ca.pz);

    ref2[0] = ref1[0];
    ref2[1] = ref1[1];
    ref2[2] = ref1[2];
    ref2[3] = -(ref2[0] * Cb.px + ref2[1] * Cb.py + ref2[2] * Cb.pz);

    /* randomly permute the constraint array */
    if (perm_timer == 0 || ALENGTH(slack) != oldslack)	{
        randperm (ALENGTH (slack), permutation);
	oldslack = ALENGTH(slack);
	}
    else perm_timer--;

    perm_slack = create_array (sizeof(PLUCKER));
    for (k = 0; k < ALENGTH(slack); k++)    {
	PP = (PLUCKER *) select_array (slack, permutation[k]);
	PP->cid = k;
	insert_array (perm_slack, (void *)PP);
	}

    len = ALENGTH(perm_slack);

    *attempts = 0;
    while (len-- > 0) {
	truncate_array (proof, 0);
	truncate_array (broken, 0);
        (*attempts)++;
        nsolns = p_underline (perm_slack, ALENGTH(perm_slack), tight, 0, L1, L2, ref1, ref2, proof, broken);
	if (nsolns > 0)
	    break;
	else	{	/* cycle */
	    truncate_array (slack, 0);
	    for (k = 0; k < ALENGTH(perm_slack); k++)
		insert_array (slack, select_array (perm_slack, k));

	    truncate_array (perm_slack, 0);
	    for (k = 0; k < ALENGTH(slack); k++)
		insert_array (perm_slack, select_array (slack, (k + 1) % ALENGTH(slack)));
	    }
	}

    discard_array (&slack);
    discard_array (&perm_slack);
    discard_array (&tight);

    if (nsolns <= 0)
	perm_timer = 16;

    return (nsolns);
}

#endif /* PUNDERLINE */

/* p_extract_constraints:
/*  given list of contours, return RANDOMIZED list of induced plucker constraints */
/*  must have called phalves = create_array (sizeof (PLUCKER)) before entry    */
int
p_extract_constraints (
    CONTOUR **contours,		    /* ith contour [x0,y0,z0; x1,y1,z1; ... xn-1, yn-1, zn-1] */
    int n,			    /* # of contours */
    ARRAY *phalves)		    /* plucker halfspaces */
{
int k, cid, pid;
POINT Ca, Cb, P_P, *P = &P_P, Q_Q, *Q = &Q_Q;
PLUCKER PL, *dp, *rp;
static ARRAY *dhalves;
static int *perm = NULL, nperm = -1;
CONTOUR *c;

    if (dhalves == NULL)
	dhalves = create_array (sizeof (PLUCKER));
    else
	truncate_array (dhalves, 0);

    /* get global contour info */
    g_contours = contours;
    g_ncontours = n;

    /* extract contour, edge id's */
    for (cid = 0; cid < n; cid++) {
	c = contours[cid];
	if (CLENGTH(c) >= 2)
            for (pid = 0; pid < CLENGTH(c); pid++)	{
		PL.uid = cid;	/* id of generator polygon */
		PL.pid = pid;
		PL.pid2 = (pid + 1) % CLENGTH(c);
		PL.special = FALSE;
		PL.cid = -1;	/* constraint id */

	        /* construct the oriented plucker line from vk to vk+1, in place */
		P = CVERTEX(contours[PL.uid], PL.pid);
		Q = CVERTEX(contours[PL.uid], PL.pid2);
	    	p_construct (P->px, P->py, P->pz, 1.0, Q->px, Q->py, Q->pz, 1.0, PL.L);
	    	p_special_normalize (PL.L, 6);

		PL.a = CVERTEX(contours[cid], PL.pid);
		PL.b = CVERTEX(contours[cid], PL.pid2);
		PL.peqn = CPLANE(contours[cid]);
		insert_array (dhalves, (void *)&PL);

	        /* special case: if this is a length 2 contour, don't enter symmetric constraint */
		if (CLENGTH(c) == 2)
		    break;
		}
	}

    /* randomize dhalves into phalves */
    if (ALENGTH(dhalves) > nperm)   {
	if (perm != NULL)
	    free (perm);
	perm = (int *) malloc((nperm = 2 * ALENGTH(dhalves)) * sizeof(int));
	}

    if (p_randomize)
        randperm (ALENGTH(dhalves), perm);
    else
        for (k = 0; k < ALENGTH(dhalves); k++)
	    perm[k] = k;		

    truncate_array (phalves, 0);
    for (k = 0; k < ALENGTH(dhalves); k++)  {
	dp = (PLUCKER *)select_array (dhalves, perm[k]);
	rp = (PLUCKER *)extend_array (phalves);
	*rp = *dp;
	assert (rp->cid == -1);
	rp->cid = k;	    /* assign constraint id */
	}

    return (ALENGTH(phalves));
}

/* returns 1 iff 5 points are in general position in 5 dimensions */
int
generalposition5d (
    float *p0, float *p1, float *p2, float *p3, float *p4)
{
int k;
DArray hdata = NULL;
float hnorm[6];
    
    if (hdata == NULL)
        hdata = (DArray) CreateDArray (6, 6);

    for (k = 0; k < 5; k++) {
	AREF(hdata, k, 0) = p0[k];
	AREF(hdata, k, 1) = p1[k];
	AREF(hdata, k, 2) = p2[k];
	AREF(hdata, k, 3) = p3[k];
	AREF(hdata, k, 4) = p4[k];

	/* last column is null */
	AREF (hdata, k, 5) = 0.0;

	/* bottom row is ones */
	AREF (hdata, 5, k) = 1.0;
	}

    AREF (hdata, 5, 5) = 0.0;
    return (Plane(5, hdata, hnorm));
}

int
intcomp (int *a, int *b)
{
    if (*a < *b)
	return (-1);
    else if (*a > *b)
	return (1);
    else
	return (0);
}

/* special hack: this is outside to allow saving the point set in wilks' format */
static ARRAY *pdualpts = NULL;
static ARRAY *puniqpts = NULL;

void
emit_segments()
{
#if 0
int k, n;
FILE *f;
char fname[128];
static int fid = 0;
float *p;

    sprintf (fname, "%s.%d", "lines", fid++);

    if ((f = fopen (fname, "w")) == NULL)
        exit (fprintf (stderr, "couldn't open %s for write\n", fname));

    fprintf (stderr, "writing %d segments to '%s'\n", nsegments, fname);

    /* fprintf (f, "%d\n", nsegments); */

    for (k = 0; k < nsegments; k++) {
        fprintf (f, "%11.8f %11.8f %11.8f\n", segments[k][0], segments[k][1], segments[k][2]); 
        fprintf (f, "%11.8f %11.8f %11.8f\n", segments[k][4], segments[k][5], segments[k][6]); 
	}

    fclose(f);
#endif
}

void
snapshot_nd_data()
{
int k;
FILE *f;
char fname[128];
static int fid = 0;
float *p;

    sprintf (fname, "%s.%d", "nd.data", fid++);

    if ((f = fopen (fname, "w")) == NULL)
        exit (fprintf (stderr, "couldn't open %s for write\n", fname));
    fprintf (stderr, "writing %d %d-d data points to '%s'\n", ALENGTH(puniqpts), 5, fname);

    fprintf (f, "%d %d\n", ALENGTH(puniqpts), 5);
    for (k = 0; k < ALENGTH(puniqpts); k++) {
	p = (float *)select_array (puniqpts, k);
        fprintf (f, "%11.7f %11.7f %11.7f %11.7f %11.7f\n", p[0], p[1], p[2], p[3], p[4]);
        }

    fclose(f);
}

typedef struct radsortstruct {
    float *dualpt;
    float  radius;
    int    source;
} RADSORT;

/* reversed to sort into DEcreasing radius */
int 
rscomp (
    RADSORT *A, RADSORT *B)
{
    if (A->radius < B->radius)
	return ( 1);	    /* was -1 */
    else if (A->radius > B->radius)
	return (-1);	    /* was  1 */
    else
	return ( 0);
}

/* returns n >= 1 and n solutions if stabbing line exists to satisfy given constraints */
/* returns -1 if halfspaces alone imply no feasible solution */
/* otherwise, returns 0 after 0(n^3) test */
/* each phalf H_i is a 6-coeff oriented hyperplane.  find a point F feasible wrt all H_i */
/* dualize the H_i w.r.t. F to form the H_i* */
int
p_stabbing_line (
    ARRAY *phalves,		/* array of PLUCKER constraints */
    ARRAY *stabbers)		/* array of stabbing lines */
{
int cid, j, k, len;

static int	*source;		/* O(ndpoints) identifiers */

static int	ndpoints = -1;

static int	nsimplices = -1;
static int	*simplices = NULL;	/* O(ndpoints) * d id's each */

static int	ndchedges;
static int	*dchedges;		/* O(ndpoints) * 2 id's each */

#ifdef PLANE_EQS
static float	*plane_eqs;		/* one per simplex, (d+1) coeffs each */
#endif

static int lastnp;

static RADSORT *radps = NULL;
static int nradps = -1;

int nd = 5;
float centroid[5];
float pF[7];		/* well-feasible point */
float rdenom, *H_i, *D_i, *P_i;
float dist;
int swap1, swap2;
int q, r, s, n_lt, n_gt, n_eq, nsolns, A, B, *sA, *sB;
int econtrib[6];
PLUCKER *l0, *l1, *l2, *l3;
float L1[6], L2[6];
int n_dupes;

    /* fix as with compute_antipen... */

#ifdef NOTDEF
    /* find a 5-D point "well-feasible" with respect to the H_i */
    /* if no such point, we can bail immediately */
    if (!interior5dA (phalves, pF)) {
	fprintf (stderr, "no interior point...\n");
	return (-1);
	}
#endif

#ifdef NORM7
    if (fabs(pF[6]) > EPS)  {
        pF[0] /= pF[6];
        pF[1] /= pF[6];
        pF[2] /= pF[6];
        pF[3] /= pF[6];
        pF[4] /= pF[6];
        pF[5] /= pF[6];
        pF[6] = 1.0;
	}
    else
	return (-1);
#else
    if (fabs(pF[5]) > EPS)  {
        pF[0] /= pF[5];
        pF[1] /= pF[5];
        pF[2] /= pF[5];
        pF[3] /= pF[5];
        pF[4] /= pF[5];
        pF[5] = 1.0;
	}
    else
	return (-1);
#endif


    fprintf (stderr, "%7.3f %7.3f %7.3f %7.3f %7.3f %7.3f %7.3f\n", 
		    pF[0], pF[1], pF[2], pF[3], pF[4], pF[5], pF[6]);

    /* dualize each halfspace to a point, and find the 5-d convex hull of the resulting set */

    /* gov up space */
    if (ALENGTH(phalves) > lastnp)   {
	if (simplices != NULL)	{
	    free (simplices);	
	    free (source);
	    free (dchedges);
#ifdef PLANE_EQS
	    free (plane_eqs);
#endif
	    }

	lastnp = 2 * MAX (8, ALENGTH(phalves));

	/* storage estimates are from edelsbrunner chapter 8, for d == 5 */

	/* O(n) storage for mapping between dual points and source halfspaces */
	source = (int *)malloc (lastnp * (nd+1) * sizeof (int));

	/* O(n^2) storage for simplices (point lists, by reference) */	
	simplices = (int *)malloc (2 * lastnp * lastnp * (nd+1) * (nd+1) * sizeof (int));

	/* O(n^2) storage for convex hull edges */
	dchedges = (int *)malloc (2 * lastnp * lastnp * (nd+1) * (nd+1) * sizeof (int));

#ifdef PLANE_EQS
	/* O(n^2) storage for plane equations (simplex normals) */
	plane_eqs = (float *)malloc (lastnp * lastnp * (nd+1) * (nd+1) * sizeof (float));

	assert (source != NULL && simplices != NULL && dchedges != NULL && plane_eqs != NULL);
#else
	assert (source != NULL && simplices != NULL && dchedges != NULL);
#endif
	}

    /* dualize each plucker halfspace with respect to pF */
    if (pdualpts == NULL)   {
        pdualpts = create_array (5 * sizeof (float));
	puniqpts = create_array (5 * sizeof (float));
	}
    else
	truncate_array (pdualpts, 0);

    for (k = 0; k < ALENGTH(phalves); k++)  {
	H_i = (float *)select_array (phalves, k);
	D_i = (float *)extend_array (pdualpts);

	/* check: is feasible point pF strictly above/below every halfspace? */
        /* if not, we should project onto an appropriate 4-D plane, and solve lower-dim problem */
	rdenom = p_fdot (pF, H_i, 5) + H_i[5];	    /* what? taking recip? */
	if (ABS(rdenom) < 1E-7)			    /* degenerate convex hull; give up for now */
	    return (0);
	rdenom = 1.0 / rdenom;
	if (ABS(rdenom) < 1E-7)			    /* unbounded convex hull; give up for now */
	    return (0);

	/* compute the dual point in 5-D */
	D_i[0] = H_i[0] * rdenom;
	D_i[1] = H_i[1] * rdenom;
	D_i[2] = H_i[2] * rdenom;
	D_i[3] = H_i[3] * rdenom;
	D_i[4] = H_i[4] * rdenom;
	}

    ndpoints = ALENGTH (pdualpts);

#define DORADSORT
#ifdef DORADSORT
    /* sort the points by distance from the origin */
    /* as they are added incrementally, all will on the convex hull */
    if (nradps < ndpoints) {
        if (radps != NULL)
	    free (radps);
        radps = (RADSORT *)malloc ((nradps = 2 * ndpoints) * sizeof (RADSORT));
        }

    for (k = 0; k < ndpoints; k++)  {
        radps[k].dualpt = (float *)select_array (pdualpts, k);
        radps[k].radius = p_dot (radps[k].dualpt, radps[k].dualpt, 5);
        radps[k].source = k;
        }

    qsort (radps, ndpoints, sizeof (RADSORT), (PF2PCVRI) rscomp);

    n_dupes = 0;
#define NOMULTIPLES
#ifdef NOMULTIPLES
		    
    /* sorted by radial distance; throw out interior duplicates */
    for (k = 1; k < ndpoints; k++)
        for (j = 0; j < k; j++)	{
            D_i = (float *)radps[k].dualpt;
	    P_i = (float *)radps[j].dualpt;

	    if (radps[j].source >= 0 && p_multiples (P_i, D_i, 5))  {
		fprintf (stderr, "dups: j %2d: %8.4f %8.4f %8.4f %8.4f %8.4f\n", j, P_i[0], P_i[1], P_i[2], P_i[3], P_i[4]);
		fprintf (stderr, "      k %2d: %8.4f %8.4f %8.4f %8.4f %8.4f\n", k, D_i[0], D_i[1], D_i[2], D_i[3], D_i[4]);
		radps[j].source = -1;
		n_dupes++;
		}
	    }

    if (n_dupes > 0)
	fprintf (stderr, "found %d duplicate points!\n", n_dupes);

#endif /* NOMULTIPLES */

    truncate_array (puniqpts, 0);
    for (j = k = 0; k < ndpoints; k++)
        if (radps[k].source >= 0)  {
	    (void) extend_array (puniqpts);
            poke_array (puniqpts, j, (void *)radps[k].dualpt);
	    source[j++] = radps[k].source;
	    }

    ndpoints -= n_dupes;
#endif

    /* check that source array is meaningful */
    for (k = 0; k < ndpoints; k++)  {
        D_i = (float *)select_array (puniqpts, k);
	assert (source[k] >= 0 && source [k] < ALENGTH(phalves));
	fprintf (stderr, "p_%2d is %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f [norm %6.3f]\n",
		k, D_i[0], D_i[1], D_i[2], D_i[3], D_i[4], D_i[5], p_dot (D_i, D_i, 5));
	}

    /* call the delaunay code [courtesy mcintosh / wilks @ att bell labs] */
    /* note NULL in plane_eqs slot; we don't care about them */
    if (p_armed)	{
	snapshot_all_data();
	p_armed = FALSE;
	}

    ndelaunay (ndpoints, nd, (float *)select_array (puniqpts, 0), FALSE /* Trace */, FALSE /* Verify */, 
	&nsimplices, simplices, &ndchedges, dchedges, NULL /* plane_eqs */);

    /* the result is a list of 5-simplices; that is, lists of 5 points, which dualize to original planes */
    /* two ADJACENT simplices share four points; these correspond to four lines each */
    /* so... since there are O(n^2) neighbor pairs, simply find the four plane equations */
    /* involved, solve for the 0,1,2, inf_1 or inf_2 lines through them, and check these */
    /* lines against all O(n) plucker constraints; this is an O(n^3) algorithm */
    /* could be made O(n^2) by checking only against a constant number (d) of plucker constraints */
    /* or, simpler: find the intersection points of the tight line(s) with the 4 source lines, */
    /* and check that these source points lie in the relative interiors of the source line segs */

    /* sort each simplex by point id */
    for (k = 0; k < nsimplices; k++)    {
	/* each simplex is a list of d+1 == 6 point id's, one of which is -1 */
	/* sort them in place with void qsort ((char *) base, nel, sizeof (*base), compar); */
	qsort ((char *)&simplices[k * (nd+1)], nd+1, sizeof (int), (PF2PCVRI) intcomp);
	assert (simplices[k * (nd+1)] == -1);
	}

    /* for each edge of the convex hull */
    for (k = 0; k < ndchedges; k++)    {
	A = dchedges[k * 2];		    /* simplex id */
	B = dchedges[k * 2 + 1];	    /* simplex id */

	/* find the d-1 == 4 points common to A & B, which are neighbor simplices */
	sA = &simplices[A * (nd+1)];
	sB = &simplices[B * (nd+1)];

	s = 0;
	econtrib[0] = econtrib[1] = econtrib[2] = econtrib[3] = -1;
	for (q = 1; q <= nd; q++)
	    for (r = 1; r <= nd; r++)	{
		if (sA[q] == sB[r]) {
		    econtrib[s++] = sA[q];
		    break;
		    }
		}

	/* must have found d-1 contributing vertices */
	/* want to assert (s == nd-1);, BUT there may be an */
	/* att bug?  sometimes sA, sB are IDENTICAL, that is, share d vertices + Infinity! */
	/* as a work around, detect this degeneracy and keep going */
    
	assert (s >= nd-1);
	if (s == nd)	{	
	    fprintf (stderr, "found identical neighbor simplices %d %d %d %d %d %d!\n", 
		sA[0], sA[1], sA[2], sA[3], sA[4], sA[5], sA[6]);
	    snapshot_all_data();
	    continue;
	    }

	/* now each point was a plucker (plane) constraint in the original list */
	/* find the 0,1,2 or infinity lines through the implicated four lines */
	l0 = (PLUCKER *)select_array (phalves, source[econtrib[0]]);
	l1 = (PLUCKER *)select_array (phalves, source[econtrib[1]]);
	l2 = (PLUCKER *)select_array (phalves, source[econtrib[2]]);
	l3 = (PLUCKER *)select_array (phalves, source[econtrib[3]]);

	/* no three of the lines can come from the same contour */
    	if (THREE_IDENTICAL (l0->uid, l1->uid, l2->uid, l3->uid))
	    continue;

	nsolns = p_incident4 (l0->L, l1->L, l2->L, l3->L, L1, L2);

	if (nsolns < 1 || nsolns > 2)
	    continue;

	/* now check the soln's against all remaining halfspaces */
	/* NYI: the case in which nsolns == 3, that is, a plane of solutions exists */
	n_lt = n_eq = n_gt = 0;
	for (q = 0; q < ALENGTH(phalves); q++)	{
	    dist = p_product (L1, (float *)select_array (phalves, q));
	    if (dist < -P_TOL)
		n_lt++;
	    else if (dist > P_TOL)
		n_gt++;	
	    else
		n_eq++;
	    if (n_lt && n_gt)
		break;
	    }

	if ((n_lt + n_eq == ALENGTH(phalves)) || (n_eq + n_gt == ALENGTH(phalves)))
	    insert_array (stabbers, (void *)L1);

	if (nsolns != 2)
	    continue;

	n_lt = n_eq = n_gt = 0;
	for (q = 0; q < ALENGTH(phalves); q++)	{
	    dist = p_product (L2, (float *)select_array (phalves, q));
	    if (dist < -P_TOL)
		n_lt++;
	    else if (dist > P_TOL)
		n_gt++;	
	    else
		n_eq++;
	    if (n_lt && n_gt)
		break;
	    }

	if ((n_lt + n_eq == ALENGTH(phalves)) || (n_eq + n_gt == ALENGTH(phalves)))
	    insert_array (stabbers, (void *)L2);

	}   /* for each dual convex hull edge */

    return (ALENGTH(stabbers));
}

#ifdef STANDALONE 
main(argc, argv)
int	argc;
char	*argv[];
{
int k, working;
double A[4], B[4], L[4][6], T1[6], T2[6];

    init_random();

    if (argc > 1)   {
	fprintf (stderr, "plucker: self-test mode\n");
	working = 1;
	for (k = 0; k < 100000; k++)  {
	    p_construct (A[0] = randr(-10.0, 10.0), 
			 A[1] = randr(-10.0, 10.0), 
			 A[2] = randr(-10.0, 10.0),
			 A[3] = randr(  0.1, 10.0),
			 randr(-10.0, 10.0),
			 randr(-10.0, 10.0),
			 randr(-10.0, 10.0),
			 randr(  0.1, 10.0),
			 L[0]);
	    p_construct (A[0],
			 A[1],
			 A[2],
			 A[3],
			 randr(-10.0, 10.0),
			 randr(-10.0, 10.0),
			 randr(-10.0, 10.0),
			 randr(  0.1, 10.0),
			 L[1]);
	    working &= p_incident (L[0], L[1]);
	    }
	fprintf (stderr, "plucker %s sanity testing\n", working ? "passed" : "failed");    
	exit (!working);
	}

    for (k = 0; k < 4; k++) {
        /* input format: one homogeneous point per line */
	if (scanf ("%lf %lf %lf %lf\n", A, A+1, A+2, A+3) != 4)
	    exit (fprintf (stderr, "expecting point A line %d\n", k));
        if (scanf ("%lf %lf %lf %lf\n", B, B+1, B+2, B+3) != 4)
	    exit (fprintf (stderr, "expecting point B line %d\n", k));
	p_construct (A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3], L[k]);
	}

    p_incident4 (L[0], L[1], L[2], L[3], T1, T2);
    
    fprintf (stderr, "sol'n 1: %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf\n", T1[0], T1[1], T1[2], T1[3], T1[4], T1[5]);
    p_deconstruct (T1, A, B);
    fprintf (stderr, "two pts: (%6.2lf %6.2lf %6.2lf %6.2lf), (%6.2lf %6.2lf %6.2lf %6.2lf)\n", 
		    A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3]);

    fprintf (stderr, "sol'n 2: %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf %6.2lf\n", T2[0], T2[1], T2[2], T2[3], T2[4], T2[5]);
    p_deconstruct (T2, A, B);
    fprintf (stderr, "two pts: (%6.2lf %6.2lf %6.2lf %6.2lf), (%6.2lf %6.2lf %6.2lf %6.2lf)\n", 
		    A[0], A[1], A[2], A[3], B[0], B[1], B[2], B[3]);
}
#endif /* STANDALONE */
